@@ -5,14 +5,14 @@ use warnings;
 
 use base qw/ Class::Accessor::Fast /;
 
-__PACKAGE__->mk_accessors(qw/ key secret error service debug /);
+__PACKAGE__->mk_accessors(qw/ key secret error status_code service debug /);
 
 use LWP::UserAgent;
 use XML::Simple;
 use URI::Escape;
 use Digest::SHA qw(hmac_sha256_base64);
 
-our $VERSION = '1.2';
+our $VERSION = '1.3';
 
 
 sub CreateTopic
@@ -39,6 +39,16 @@ sub GetTopic
 	});
 }
 
+sub GetTarget
+{
+	my ($self, $arn) = @_;
+
+	return Amazon::SNS::Target->new({
+		'sns' => $self,
+		'arn' => $arn,
+	});
+}
+
 sub DeleteTopic
 {
 	my ($self, $arn) = @_;
@@ -67,6 +77,27 @@ sub ListTopics
 	} @{$r->{'ListTopicsResult'}{'Topics'}[0]{'member'}};
 }
 
+sub Subscribe
+{
+	my ($self, $protocol, $topicarn, $endpoint) = @_;
+
+	$self->dispatch({
+		'Action'	=> 'Subscribe',
+		'Protocol'	=> $protocol,
+		'TopicArn'	=> $topicarn,
+		'Endpoint'	=> $endpoint,
+	});
+}
+
+sub Unsubscribe
+{
+	my ($self, $arn) = @_;
+
+	$self->dispatch({
+		'Action'		=> 'Unsubscribe',
+		'SubscriptionArn'	=> $arn,
+	});
+}
 
 sub dispatch
 {
@@ -85,6 +116,14 @@ sub dispatch
 	$args->{'AWSAccessKeyId'} = $self->key;
 	$args->{'SignatureVersion'} = 2;
 	$args->{'SignatureMethod'} = 'HmacSHA256';
+	$args->{'Version'} = '2010-03-31';
+
+	if (defined($args->{'Attributes'}) and ref($args->{'Attributes'}) eq 'HASH') {
+		foreach my $attr (keys %{$args->{'Attributes'}}) {
+			$args->{$attr} = $args->{'Attributes'}->{$attr};
+		}
+		delete $args->{'Attributes'};
+	}
 
 	# build URI
 	my $uri = URI->new($self->service);
@@ -105,20 +144,22 @@ sub dispatch
 
 	my $response = LWP::UserAgent->new->post($self->service, 'Content' => $uri->query);
 
-        if ($response->is_success) {
+	$self->status_code($response->code);
+
+	if ($response->is_success) {
 		return XMLin($response->content,
 				'SuppressEmpty' => 1,
 #				'KeyAttr' => { },
 				'ForceArray' => [ qw/ Topics member / ],
 		);
-        } else {
+	} else {
 		print $response->content, "\n";
 		$self->error(
 			($response->content =~ /^<.+>/)
 				? eval { XMLin($response->content)->{'Error'}{'Message'} || $response->status_line }
 				: $response->status_line
 		);
-        }
+	}
 
 	print STDERR 'ERROR: ', $self->error, "\n"
 		if $self->debug;
@@ -184,6 +225,69 @@ sub DeleteTopic
 
 1;
 
+
+package Amazon::SNS::Target;
+
+use strict;
+use warnings;
+
+use base qw(Class::Accessor);
+
+use JSON;
+
+__PACKAGE__->mk_accessors(qw/ sns arn /);
+
+sub Publish
+{
+	my ($self, $msg, $subj, $attr) = @_;
+
+	# XXX croak on invalid arn
+
+	my $structure = undef;
+	my $attributes = undef;
+
+	# support JSON payload
+	if (ref($msg) eq 'HASH') {
+
+		$structure = 'json';
+		$msg = encode_json($msg);
+	}
+
+	if (defined($attr) and ref($attr) eq 'HASH') {
+
+		my $i = 1;
+
+		foreach my $key (keys %$attr) {
+
+			$attributes->{"MessageAttributes.entry.$i.Name"} = $key;
+			$attributes->{"MessageAttributes.entry.$i.Value.DataType"} = $attr->{$key}->{'Type'};
+
+			if($attr->{$key}->{'Type'} eq 'Binary') {
+				$attributes->{"MessageAttributes.entry.$i.Value.BinaryValue"} = $attr->{$key}->{'Value'};
+			} else {
+				$attributes->{"MessageAttributes.entry.$i.Value.StringValue"} = $attr->{$key}->{'Value'};
+			}
+
+			$i++;
+		}
+	}
+
+	my $r = $self->sns->dispatch({
+		'Action'		=> 'Publish',
+		'TargetArn'		=> $self->arn,
+		'Message'		=> $msg,
+		'MessageStructure'	=> $structure,
+		'Subject'		=> $subj,
+		'Attributes'		=> $attributes,
+	});
+
+	# return message id on success, undef on error
+	return $r ? $r->{'PublishResult'}{'MessageId'} : undef;
+}
+
+1;
+
+
 =head1 NAME
 
 Amazon::SNS - Amazon Simple Notification Service made simpler
@@ -197,7 +301,7 @@ Amazon::SNS - Amazon Simple Notification Service made simpler
 
   # create a new topic and publish
 
-  my $topic = $sns->CeateTopic('MyTopic')
+  my $topic = $sns->CreateTopic('MyTopic')
 	or die $sns->error;
 
   $topic->Publish('My test message');
@@ -212,7 +316,7 @@ Amazon::SNS - Amazon Simple Notification Service made simpler
 
   my $topic = $sns->GetTopic('arn:aws:sns:eu-west-1:123456789099:MyTopic');
 
-  $topic->Publish('My test message');
+  $topic->Publish('My test message', 'My Subject');
 
 
   # get all topics
@@ -243,6 +347,16 @@ Sorry for not providing a better documentation, patches are always accepted. ;)
 
 	Gives you an Amazon::SNS::Topic object using an existing ARN.
 
+=item $sns->GetTarget($arn)
+
+	Gives you an Amazon::SNS::Target object using an existing ARN. Sending Notification to TargetArn instead of TopicArn.
+
+=item $sns->Publish($message, $subject, $attributes) (Amazon::SNS::Target)
+
+	When used with Amazon::SNS::Target object (see GetTarget), additional parameter $attributes is used to pass MessageAttributes.entry.N attributes with message.
+	An example of MobilePush TTL: $attributes = {"AWS.SNS.MOBILE.APNS.TTL" => {"Type" => "String", "Value" => 3600}};
+	More information can be found on Amazon web site: http://docs.aws.amazon.com/sns/latest/dg/sns-ttl.html
+
 =item $sns->CreateTopic($name)
 
 	Gives you an Amazon::SNS::Topic object with the given name, creating it 
@@ -260,6 +374,10 @@ Sorry for not providing a better documentation, patches are always accepted. ;)
 
 	Description of the last error, or undef if none.
 
+=item $sns->status_code
+
+	The status code of the last HTTP response.
+
 =back
 
 =head1 ATTRIBUTES
@@ -305,7 +423,7 @@ Alessandro Zummo, E<lt>a.zummo@towertech.itE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2011 by Alessandro Zummo
+Copyright (C) 2011-15 Alessandro Zummo
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License version 2 as
